/*-
 *  Copyright 2020 Beuth Hochschule für Technik Berlin, Hochschule für Technik Stuttgart
 *
 *  This file is part of CityDoctor2.
 *
 *  CityDoctor2 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  CityDoctor2 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with CityDoctor2.  If not, see <https://www.gnu.org/licenses/>.
 */
package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.citygml4j.core.model.core.AbstractSpaceBoundaryProperty;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurface;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurfaceProperty;
import org.xmlobjects.gml.model.geometry.primitives.Solid;
import org.xmlobjects.gml.model.geometry.primitives.SolidProperty;

import java.io.Serial;
import java.util.ArrayList;
import java.util.List;

/**
 * Container for attributes of an abstract building. It can be a Building or
 * BuildingPart.
 *
 * @author Matthias Betz
 */
public abstract class AbstractBuilding extends CityObject {

    @Serial
    private static final long serialVersionUID = -2196414503088741206L;

    private static final Logger logger = LogManager.getLogger(AbstractBuilding.class);

    private final List<Installation> buildingInstallations = new ArrayList<>(2);
    private final List<BoundarySurface> boundarySurfaceList = new ArrayList<>();
    private final List<BuildingRoom> buildingRooms = new ArrayList<>();
    private final List<Storey> buildingStoreys = new ArrayList<>();
    private final List<BuildingUnit> buildingUnits = new ArrayList<>();
    private final List<BuildingRoomFurniture> buildingRoomFurnitureList = new ArrayList<>();
    private org.citygml4j.core.model.building.AbstractBuilding ab;

    /**
     * Getter for all boundary surfaces contained in this building.
     *
     * @return the boundary surfaces
     */
    public List<BoundarySurface> getBoundarySurfaces() {
        return boundarySurfaceList;
    }

    @Override
    public org.citygml4j.core.model.building.AbstractBuilding getGmlObject() {
        return ab;
    }

    @Override
    public FeatureType getFeatureType() {
        return FeatureType.BUILDING;
    }

    @Override
    public void unsetGmlGeometries() {
        ab.setLod1Solid(null);
        ab.setLod2Solid(null);
        ab.setLod3Solid(null);
        ab.setLod2MultiSurface(null);
        ab.setLod3MultiSurface(null);
        ab.getDeprecatedProperties().setLod1MultiSurface(null);
        ab.getDeprecatedProperties().setLod4MultiSurface(null);
        ab.getDeprecatedProperties().setLod4Solid(null);

        for (BoundarySurface bs : boundarySurfaceList) {
            bs.unsetGmlGeometries();
        }
        for (Installation bi : buildingInstallations) {
            bi.unsetGmlGeometries();
        }
        for (BuildingRoom br : buildingRooms) {
            br.unsetGmlGeometries();
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            bfr.unsetGmlGeometries();
        }
        for (Storey storey : buildingStoreys) {
            storey.unsetGmlGeometries();
        }
        for (BuildingUnit bu : buildingUnits) {
            bu.unsetGmlGeometries();
        }
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        for (Geometry geom : getGeometries()) {
            if (geom instanceof ImplicitGeometryHolder) {
                continue;
            }
            if (geom.getType() == GeometryType.MULTI_SURFACE) {
                MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
                setMultiSurfaceAccordingToLod(geom, ms);
            } else {
                Solid solid = CityGmlUtils.createSolid(geom, factory, config);
                setSolidAccordingToLod(geom, solid);
            }
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            reCreateBoundarySurface(factory, config, bs);
        }
        for (Installation bi : buildingInstallations) {
            bi.reCreateGeometries(factory, config);
        }
        for (BuildingRoom br : buildingRooms) {
            br.reCreateGeometries(factory, config);
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            bfr.reCreateGeometries(factory, config);
        }
        for (Storey storey : buildingStoreys) {
            storey.reCreateGeometries(factory, config);
        }
        for (BuildingUnit bu : buildingUnits) {
            bu.reCreateGeometries(factory, config);
        }
    }

    private void reCreateBoundarySurface(GeometryFactory factory, ParserConfiguration config, BoundarySurface bs) {
        if (bs.getGeometries().isEmpty()) {
            for (AbstractSpaceBoundaryProperty bsp : ab.getBoundaries()) {
                if (bsp.getObject() != null && bsp.getObject() == bs.getGmlObject()) {
                    logger.warn("Found empty boundary surface: {}, removing from building", bs.getGmlId());
                    ab.getBoundaries().remove(bsp);
                    break;
                }
            }
            return;
        }
        bs.reCreateGeometries(factory, config);
    }

    private void setMultiSurfaceAccordingToLod(Geometry geom, MultiSurface ms) {
        switch (geom.getLod()) {
            case LOD0:
                ab.setLod0MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD1:
                ab.getDeprecatedProperties().setLod1MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD2:
                ab.setLod2MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD3:
                ab.setLod3MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD4:
                ab.getDeprecatedProperties().setLod4MultiSurface(new MultiSurfaceProperty(ms));
                break;
            default:
                throw new IllegalStateException("Cannot add " + geom.getLod() + " multi surface to buildings");
        }
    }

    private void setSolidAccordingToLod(Geometry geom, Solid solid) {
        switch (geom.getLod()) {
            case LOD1:
                ab.setLod1Solid(new SolidProperty(solid));
                break;
            case LOD2:
                ab.setLod2Solid(new SolidProperty(solid));
                break;
            case LOD3:
                ab.setLod3Solid(new SolidProperty(solid));
                break;
            case LOD4:
                ab.getDeprecatedProperties().setLod4Solid(new SolidProperty(solid));
                break;
            default:
                throw new IllegalStateException("Cannot add " + geom.getLod() + " solid to buildings");
        }
    }

    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (Installation bi : buildingInstallations) {
            bi.accept(c);
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.accept(c);
        }
        for (BuildingRoom br : buildingRooms) {
            br.accept(c);
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            bfr.accept(c);
        }
        for (Storey storey : buildingStoreys) {
            storey.accept(c);
        }
        for (BuildingUnit bu : buildingUnits) {
            bu.accept(c);
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (Installation bi : buildingInstallations) {
            bi.collectContainedErrors(errors);
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.collectContainedErrors(errors);
        }
        for (BuildingRoom br : buildingRooms) {
            br.collectContainedErrors(errors);
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            bfr.collectContainedErrors(errors);
        }
        for (Storey storey : buildingStoreys) {
            storey.collectContainedErrors(errors);
        }
        for (BuildingUnit bu : buildingUnits) {
            bu.collectContainedErrors(errors);
        }
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (Installation bi : buildingInstallations) {
            bi.clearAllContainedCheckResults();
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.clearAllContainedCheckResults();
        }
        for (BuildingRoom br : buildingRooms) {
            br.clearAllContainedCheckResults();
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            bfr.clearAllContainedCheckResults();
        }
        for (Storey storey : buildingStoreys) {
            storey.clearAllContainedCheckResults();
        }
        for (BuildingUnit bu : buildingUnits) {
            bu.clearAllContainedCheckResults();
        }
    }

    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (Installation bi : buildingInstallations) {
            if (bi.containsError(checkIdentifier)) {
                return true;
            }
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            if (bs.containsError(checkIdentifier)) {
                return true;
            }
        }
        for (BuildingRoom br : buildingRooms) {
            if (br.containsError(checkIdentifier)) {
                return true;
            }
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            if (bfr.containsError(checkIdentifier)) {
                return true;
            }
        }
        for (Storey storey : buildingStoreys) {
            if (storey.containsError(checkIdentifier)) {
                return true;
            }
        }
        for (BuildingUnit bu : buildingUnits) {
            if (bu.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (Installation bi : buildingInstallations) {
            if (bi.containsAnyError()) {
                return true;
            }
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            if (bs.containsAnyError()) {
                return true;
            }
        }
        for (BuildingRoom br : buildingRooms) {
            if (br.containsAnyError()) {
                return true;
            }
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            if (bfr.containsAnyError()) {
                return true;
            }
        }
        for (Storey storey : buildingStoreys) {
            if (storey.containsAnyError()) {
                return true;
            }
        }
        for (BuildingUnit bu : buildingUnits) {
            if (bu.containsAnyError()) {
                return true;
            }
        }
        return false;
    }

    void setCityGmlBuilding(org.citygml4j.core.model.building.AbstractBuilding ab) {
        this.ab = ab;
    }

    public void addBoundarySurface(BoundarySurface bs) {
        boundarySurfaceList.add(bs);
        bs.setParent(this);
    }

    public void addBuildingInstallation(Installation coBi) {
        buildingInstallations.add(coBi);
        coBi.setParent(this);
    }

    public void addBuildingRoom(BuildingRoom room) {
        buildingRooms.add(room);
        room.setParent(this);
    }

    public void addBuildingRoomFurniture(BuildingRoomFurniture roomFurniture) {
        buildingRoomFurnitureList.add(roomFurniture);
        roomFurniture.setParent(this);
    }

    public void addStorey(Storey storey) {
        buildingStoreys.add(storey);
    }

    public void addBuildingUnit(BuildingUnit buildingUnit) {
        buildingUnits.add(buildingUnit);
    }

    public void setGmlObject(org.citygml4j.core.model.building.AbstractBuilding ab) {
        this.ab = ab;
    }

    public List<Installation> getBuildingInstallations() {
        return buildingInstallations;
    }

    public List<BuildingRoom> getBuildingRooms() {
        return buildingRooms;
    }

    public List<BuildingRoomFurniture> getBuildingRoomFurnitureList() {
        return buildingRoomFurnitureList;
    }

    public List<Storey> getBuildingStoreys() {
        return buildingStoreys;
    }

    public List<BuildingUnit> getBuildingUnits() {
        return buildingUnits;
    }

    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (Installation bi : buildingInstallations) {
            bi.prepareForChecking();
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.prepareForChecking();
        }
        for (BuildingRoom br : buildingRooms) {
            br.prepareForChecking();
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            bfr.prepareForChecking();
        }
        for (Storey storey : buildingStoreys) {
            storey.prepareForChecking();
        }
        for (BuildingUnit bu : buildingUnits) {
            bu.prepareForChecking();
        }
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (Installation bi : buildingInstallations) {
            bi.clearMetaInformation();
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.clearMetaInformation();
        }
        for (BuildingRoom br : buildingRooms) {
            br.clearMetaInformation();
        }
        for (BuildingRoomFurniture bfr : buildingRoomFurnitureList) {
            bfr.clearMetaInformation();
        }
        for (Storey storey : buildingStoreys) {
            storey.clearMetaInformation();
        }
        for (BuildingUnit bu : buildingUnits) {
            bu.clearMetaInformation();
        }
    }

    @Override
    public void collectInstances(CopyHandler handler) {
        super.collectInstances(handler);
        handler.addInstance(boundarySurfaceList);
        handler.addInstance(buildingInstallations);
        handler.addInstance(buildingRooms);
        handler.addInstance(buildingRoomFurnitureList);
        handler.addInstance(buildingStoreys);
        handler.addInstance(buildingUnits);
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        AbstractBuilding originalAb = (AbstractBuilding) original;
        for (BoundarySurface originalBs : originalAb.boundarySurfaceList) {
            boundarySurfaceList.add(handler.getCopyInstance(originalBs));
        }
        for (Installation originalBi : originalAb.buildingInstallations) {
            buildingInstallations.add(handler.getCopyInstance(originalBi));
        }
        for (BuildingRoom originalBr : originalAb.buildingRooms) {
            buildingRooms.add(handler.getCopyInstance(originalBr));
        }
        for (BuildingRoomFurniture originalBFR : originalAb.buildingRoomFurnitureList) {
            buildingRoomFurnitureList.add(handler.getCopyInstance(originalBFR));
        }
        for (Storey originalBStoreys : originalAb.buildingStoreys) {
            buildingStoreys.add(handler.getCopyInstance(originalBStoreys));
        }
        for (BuildingUnit originalBun : originalAb.buildingUnits) {
            buildingUnits.add(handler.getCopyInstance(originalBun));
        }
        ab = originalAb.ab;
    }

}
