package de.hft.stuttgart.citydoctor2.gui;

import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.Checkable;
import de.hft.stuttgart.citydoctor2.datastructure.*;
import de.hft.stuttgart.citydoctor2.gui.filter.ViewFilter;
import de.hft.stuttgart.citydoctor2.gui.tree.*;
import de.hft.stuttgart.citydoctor2.math.Triangle3d;
import javafx.application.Platform;
import javafx.scene.control.TreeItem;
import javafx.scene.paint.Color;
import javafx.scene.shape.CullFace;
import javafx.scene.shape.DrawMode;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.IOException;
import java.util.*;

public class Renderer {

    private static final Logger logger = LogManager.getLogger(Renderer.class);

    private TriangulatedGeometry currentTriGeom;
    private Geometry currentGeometry;

    private CullFace currentCulling = CullFace.BACK;
    private DrawMode currentDrawMode = DrawMode.FILL;

    private MainWindow mainWindow;
    private HighlightController highlightController;
    private ListErrorVisitor errVisitor;

    private LoadingInfoDialog loadingDialog;

    private List<ViewFilter> lodFilters = new ArrayList<>();
    private ViewFilter roofFilter;
    private Runnable refresher;
    private Runnable errorUpdater;

    public Renderer(MainWindow mainWindow, HighlightController highlightController) throws IOException {
        this.mainWindow = mainWindow;
        this.highlightController = highlightController;
        loadingDialog = new LoadingInfoDialog(mainWindow.getMainStage());
        errVisitor = new ListErrorVisitor(highlightController);
        setupLodFilters();
    }

    private void setupLodFilters() {
        lodFilters.add(new ViewFilter() {

            @Override
            public boolean useGeometry(CityObject co, Geometry geom) {
                return geom.getLod() == Lod.LOD1;
            }

        });
        lodFilters.add(new ViewFilter() {

            @Override
            public boolean useGeometry(CityObject co, Geometry geom) {
                return geom.getLod() == Lod.LOD2;
            }

        });
        lodFilters.add(new ViewFilter() {

            @Override
            public boolean useGeometry(CityObject co, Geometry geom) {
                return geom.getLod() == Lod.LOD3;
            }

        });
        lodFilters.add(new ViewFilter() {

            @Override
            protected boolean useGeometry(CityObject co, Geometry geom) {
                return geom.getLod() == Lod.LOD0;
            }

        });
        lodFilters.add(new ViewFilter() {

            @Override
            public boolean useGeometry(CityObject co, Geometry geom) {
                return geom.getLod() == Lod.LOD4;
            }
        });
        roofFilter = new ViewFilter() {
            @Override
            public boolean allowedToUse(CityObject co, Geometry geom) {
                if (!this.isEnabled()) {
                    return true;
                }
                return this.useGeometry(co, geom);
            }

            @Override
            public boolean useGeometry(CityObject co, Geometry geom) {
                if (co instanceof BoundarySurface bs) {
                    return bs.getType() != BoundarySurfaceType.ROOF;
                }
                return true;
            }

        };
        roofFilter.setEnable(false);
    }

    public void enableLod1() {
        lodFilters.get(0).enable();
        if (refresher != null) {
            refresher.run();
        }
    }

    public void disableLod1() {
        lodFilters.get(0).disable();
        if (refresher != null) {
            refresher.run();
        }
    }

    public void enableLod2() {
        lodFilters.get(1).enable();
        if (refresher != null) {
            refresher.run();
        }
    }

    public void disableLod2() {
        lodFilters.get(1).disable();
        if (refresher != null) {
            refresher.run();
        }
    }

    public void enableLod3() {
        lodFilters.get(2).enable();
        if (refresher != null) {
            refresher.run();
        }
    }

    public void disableLod3() {
        lodFilters.get(2).disable();
        if (refresher != null) {
            refresher.run();
        }
    }

    public void enableLod4() {
        lodFilters.get(3).enable();
        if (refresher != null) {
            refresher.run();
        }
    }

    public void disableLod4() {
        lodFilters.get(3).disable();
        if (refresher != null) {
            refresher.run();
        }
    }

    public void hideRoofs() {
        roofFilter.setEnable(true);
        if (refresher != null) {
            refresher.run();
        }
    }

    public void showRoofs() {
        roofFilter.setEnable(false);
        if (refresher != null) {
            refresher.run();
        }
    }

    public void render(Building building) {
        refresher = () -> {
            Set<ConcretePolygon> setupBuildingPolygons = setupBuildingPolygons(building);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupBuildingPolygons));
            render(setupBuildingPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(building);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupBuildingPolygons(Building b) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(b, polygons);
        for (BoundarySurface bs : b.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        for (Installation bi : b.getBuildingInstallations()) {
            addPolygons(bi, polygons);
            for (BoundarySurface bs : bi.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
                for (Opening op : bs.getOpenings()) {
                    addPolygons(op, polygons);
                }
            }
        }
        for (BuildingRoom br : b.getBuildingRooms()) {
            addPolygons(br, polygons);
            for (BoundarySurface bs : br.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
            }
        }
        for (BuildingRoomFurniture brf : b.getBuildingRoomFurnitureList()) {
            addPolygons(brf, polygons);
            for (BoundarySurface bs : brf.getBoundarySurfaceList()) {
                addPolygons(bs, polygons);
            }
        }
        for (BuildingPart bp : b.getBuildingParts()) {
            polygons.addAll(setupBuildingPartPolygons(bp));
        }
        return polygons;
    }

    public void render(BuildingPart bp) {
        refresher = () -> {
            Set<ConcretePolygon> setupBuildingPartPolygons = setupBuildingPartPolygons(bp);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupBuildingPartPolygons));
            render(setupBuildingPartPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(bp);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupBuildingPartPolygons(BuildingPart bp) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(bp, polygons);
        for (BoundarySurface bs : bp.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        for (Installation bi : bp.getBuildingInstallations()) {
            addPolygons(bi, polygons);
            for (BoundarySurface bs : bi.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
                for (Opening op : bs.getOpenings()) {
                    addPolygons(op, polygons);
                }
            }
        }
        return polygons;
    }

    public void render(AbstractRoom room) {
        refresher = () -> {
            Set<ConcretePolygon> setupRoomPolygons = setupRoomPolygons(room);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupRoomPolygons));
            render(setupRoomPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(room);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupRoomPolygons(AbstractRoom room) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(room, polygons);
        for (BoundarySurface bs : room.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        return polygons;
    }

    public void render(AbstractFurniture furniture) {
        refresher = () -> {
            Set<ConcretePolygon> setupFurniturePolygons = setupFurniturePolygons(furniture);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupFurniturePolygons));
            render(setupFurniturePolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(furniture);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupFurniturePolygons(AbstractFurniture furniture) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(furniture, polygons);
        for (BoundarySurface bs : furniture.getBoundarySurfaceList()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        return polygons;
    }

    public void render(Storey storey) {
        refresher = () -> {
            Set<ConcretePolygon> setupStoreyPolygons = setupStoreyPolygons(storey);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupStoreyPolygons));
            render(setupStoreyPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(storey);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupStoreyPolygons(Storey storey) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(storey, polygons);
        for (BoundarySurface bs : storey.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        return polygons;
    }

    public void render(BuildingUnit buildingUnit) {
        refresher = () -> {
            Set<ConcretePolygon> setupBuildingUnitPolgons = setupBuildingUnitPolygons(buildingUnit);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupBuildingUnitPolgons));
            render(setupBuildingUnitPolgons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(buildingUnit);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupBuildingUnitPolygons(BuildingUnit buildingUnit) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(buildingUnit, polygons);
        for (BoundarySurface bs : buildingUnit.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        return polygons;
    }

    public void render(Tunnel tunnel) {
        refresher = () -> {
            Set<ConcretePolygon> setupTunnelPolygons = setupTunnelPolygons(tunnel);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupTunnelPolygons));
            render(setupTunnelPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(tunnel);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupTunnelPolygons(Tunnel tunnel) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(tunnel, polygons);
        for (BoundarySurface bs : tunnel.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
        }
        for (Installation ti : tunnel.getTunnelInstallations()) {
            addPolygons(ti, polygons);
            for (BoundarySurface bs : ti.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
                for (Opening op : bs.getOpenings()) {
                    addPolygons(op, polygons);
                }
            }
        }
        for (TunnelHollow th : tunnel.getTunnelHollows()) {
            addPolygons(th, polygons);
            for (BoundarySurface bs : th.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
            }
        }
        for (TunnelFurniture tf : tunnel.getTunnelFurnitureList()) {
            addPolygons(tf, polygons);
            for (BoundarySurface bs : tf.getBoundarySurfaceList()) {
                addPolygons(bs, polygons);
            }
        }
        for (TunnelPart tp : tunnel.getTunnelParts()) {
            polygons.addAll(setupTunnelPartPolygons(tp));
        }
        return polygons;
    }

    private Set<ConcretePolygon> setupTunnelPartPolygons(TunnelPart tp) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(tp, polygons);
        for (BoundarySurface bs : tp.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        for (Installation bi : tp.getTunnelInstallations()) {
            addPolygons(bi, polygons);
            for (BoundarySurface bs : bi.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
                for (Opening op : bs.getOpenings()) {
                    addPolygons(op, polygons);
                }
            }
        }
        return polygons;
    }

    public void render(BridgeObject bridge) {
        refresher = () -> {
            Set<ConcretePolygon> setupBridgePolygons = setupBridgePolygons(bridge);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupBridgePolygons));
            render(setupBridgePolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(bridge);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupBridgePolygons(BridgeObject bridge) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(bridge, polygons);
        for (BoundarySurface bs : bridge.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
        }
        for (BridgeConstructiveElement consElement : bridge.getConstructiveElements()) {
            addPolygons(consElement, polygons);
            for (BoundarySurface bs : consElement.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
            }
        }
        for (Installation inst : bridge.getBridgeInstallations()) {
            addPolygons(inst, polygons);
            for (BoundarySurface bs : inst.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
            }
        }
        return polygons;
    }

    public void render(CityObject co) {
        refresher = () -> {
            Set<ConcretePolygon> setupCityObjectPolygons = setupCityObjectPolygons(co);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupCityObjectPolygons));
            render(setupCityObjectPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(co);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    public void render(ReliefObject relief) {
        refresher = () -> {
            Set<ConcretePolygon> setupCityObjectPolygons = setupReliefPolygons(relief);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupCityObjectPolygons));
            render(setupCityObjectPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(relief);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupReliefPolygons(ReliefObject relief) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(relief, polygons);
        for (TinObject tin : relief.getComponents()) {
            addPolygons(tin, polygons);
        }
        return polygons;
    }

    public void render(Installation bi) {
        refresher = () -> {
            Set<ConcretePolygon> setupCityObjectPolygons = setupBuildingInstallationPolygons(bi);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupCityObjectPolygons));
            render(setupCityObjectPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(bi);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupBuildingInstallationPolygons(Installation bi) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(bi, polygons);
        for (BoundarySurface bs : bi.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        return polygons;
    }

    public void render(CityFurniture cf) {
        refresher = () -> {
            Set<ConcretePolygon> setupCityFurniturePolygons = setupCityFurniturePolygons(cf);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupCityFurniturePolygons));
            render(setupCityFurniturePolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(cf);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupCityFurniturePolygons(CityFurniture cf) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(cf, polygons);
        return polygons;
    }

    public void render(GenericCityObject gco) {
        refresher = () -> {
            Set<ConcretePolygon> setupGenericCityObjectPolygons = setupGenericCityObjectPolygons(gco);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupGenericCityObjectPolygons));
            render(setupGenericCityObjectPolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(gco);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupGenericCityObjectPolygons(GenericCityObject gco) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(gco, polygons);
        for (BoundarySurface bs : gco.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening op : bs.getOpenings()) {
                addPolygons(op, polygons);
            }
        }
        return polygons;
    }

    public void render(BoundarySurface bs) {
        refresher = () -> {
            Set<ConcretePolygon> setupBoundarySurfacePolygons = setupBoundarySurfacePolygons(bs);
            mainWindow.zoomOutForBoundingBox(BoundingBox.of(setupBoundarySurfacePolygons));
            render(setupBoundarySurfacePolygons);
            Platform.runLater(() -> {
                errorUpdater = () -> displayErrors(bs);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private Set<ConcretePolygon> setupBoundarySurfacePolygons(BoundarySurface bs) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(bs, polygons);
        for (Opening op : bs.getOpenings()) {
            addPolygons(op, polygons);
        }
        return polygons;
    }

    private Set<ConcretePolygon> setupCityObjectPolygons(CityObject co) {
        Set<ConcretePolygon> polygons = new HashSet<>();
        addPolygons(co, polygons);
        return polygons;
    }

    private void addPolygons(CityObject co, Set<ConcretePolygon> polygons) {
        for (Geometry geom : co.getGeometries()) {
            boolean used = false;
            for (ViewFilter filter : lodFilters) {
                if (filter.allowedToUse(co, geom)) {
                    used = true;
                    break;
                }
            }
            if (used) {
                if (roofFilter.allowedToUse(co, geom)) {
                    addConcretePolygons(polygons, geom);
                }
            }
        }
    }

    private void displayErrors(Checkable c) {
        if (!c.isValidated()) {
            return;
        }
        List<CheckError> errors = new ArrayList<>();
        c.collectContainedErrors(errors);

        // filter out duplicate errors (polygon can be contained in multiple geometries)
        Set<CheckError> errorSet = new HashSet<>(errors);
        for (CheckError err : errorSet) {
            ErrorNode node = new ErrorNode(err);
            TreeItem<Renderable> errItem = new TreeItem<>(node);
            ErrorItemVisitor visitor = new ErrorItemVisitor(errItem);
            err.accept(visitor);
            mainWindow.getErrorTree().getRoot().getChildren().add(errItem);

        }
    }

    public void render(Geometry geom) {
        refresher = () -> {
            Platform.runLater(this::clearGeometryTrees);
            currentTriGeom = TriangulatedGeometry.of(geom);
            if (geom.getEdges() == null && currentGeometry != null) {
                // if there are no edges available low memory mode is enabled
                // clear the old geometry of all meta information
                currentGeometry.clearMetaInformation();
            }
            errVisitor.setGeometry(currentTriGeom);
            Platform.runLater(() -> {
                setupRenderState();
                if (geom.getEdges() == null) {
                    // create edges and vertices so they can be listed in the gui
                    geom.prepareForChecking();
                    // remember the geometry, so it can be cleared if another is displayed
                    currentGeometry = geom;
                }
                addGeometryDataToView(geom);
                mainWindow.zoomOutForBoundingBox(BoundingBox.of(geom.getPolygons()));
                errorUpdater = () -> displayErrors(geom);
                errorUpdater.run();
            });
        };
        refresher.run();
    }

    private void addGeometryDataToView(Geometry geom) {
        for (Polygon p : geom.getPolygons()) {
            if (p instanceof CompositePolygon) {
                continue;
            }
            addPolygonToView(p);
        }
        for (Edge e : geom.getEdges()) {
            addEdgeToView(e);
        }
        for (Vertex v : geom.getVertices()) {
            addVertexToView(v);
        }
    }

    private void addVertexToView(Vertex v) {
        Renderable cf = new VertexNode(v);
        TreeItem<Renderable> ti = new TreeItem<>(cf);
        mainWindow.getVertexView().getRoot().getChildren().add(ti);
    }

    private void addEdgeToView(Edge e) {
        EdgeNode cf = new EdgeNode(e);
        TreeItem<Renderable> ti = new TreeItem<>(cf);
        mainWindow.getEdgeView().getRoot().getChildren().add(ti);
    }

    private void addPolygonToView(Polygon p) {
        CheckStatus cs = determineCheckStatus(p);
        PolygonNode cf = new PolygonNode(p, cs);
        TreeItem<Renderable> ti = new TreeItem<>(cf);
        mainWindow.getPolygonsView().getRoot().getChildren().add(ti);

        // add linear rings
        CheckStatus csRing = determineCheckStatus(p.getExteriorRing());
        Renderable ccRing = new LinearRingNode(p.getExteriorRing(), csRing);
        TreeItem<Renderable> tiRing = new TreeItem<>(ccRing);
        ti.getChildren().add(tiRing);

        for (Vertex v : p.getExteriorRing().getVertices()) {
            VertexNode vn = new VertexNode(v);
            TreeItem<Renderable> tiV = new TreeItem<>(vn);
            tiRing.getChildren().add(tiV);
        }

        for (LinearRing lr : p.getInnerRings()) {
            CheckStatus csInteriorRing = determineCheckStatus(lr);
            Renderable ccInteriorRing = new LinearRingNode(lr, csInteriorRing);
            TreeItem<Renderable> tiInteriorRing = new TreeItem<>(ccInteriorRing);
            ti.getChildren().add(tiInteriorRing);
            for (Vertex v : lr.getVertices()) {
                VertexNode vn = new VertexNode(v);
                TreeItem<Renderable> tiV = new TreeItem<>(vn);
                tiRing.getChildren().add(tiV);
            }
        }
    }

    private CheckStatus determineCheckStatus(Checkable c) {
        if (!c.isValidated()) {
            return CheckStatus.NOT_CHECKED;
        }
        if (c.containsAnyError()) {
            return CheckStatus.ERROR;
        }
        return CheckStatus.OK;
    }

    private void render(Collection<? extends Polygon> polygons) {
        Platform.runLater(this::clearGeometryTrees);
        currentTriGeom = TriangulatedGeometry.of(polygons);
        errVisitor.setGeometry(currentTriGeom);
        setupRenderState();
    }

    private void clearGeometryTrees() {
        highlightController.clearHighlights();
        mainWindow.getErrorTree().getRoot().getChildren().clear();
        mainWindow.getPolygonsView().getRoot().getChildren().clear();
        mainWindow.getEdgeView().getRoot().getChildren().clear();
        mainWindow.getVertexView().getRoot().getChildren().clear();
    }

    public void renderBuildings(List<Building> objects) {
        errorUpdater = null;
        refresher = () -> {
            Platform.runLater(() -> {
                loadingDialog.show();
                clearGeometryTrees();
            });
            Thread t = new Thread(() -> {
                Set<ConcretePolygon> polygons = new HashSet<>();
                for (Building b : objects) {
                    collectBuildingPolygons(polygons, b);
                    for (BuildingPart bp : b.getBuildingParts()) {
                        collectBuildingPolygons(polygons, bp);
                    }
                }
                currentTriGeom = TriangulatedGeometry.of(polygons);
                errVisitor.setGeometry(currentTriGeom);
                Platform.runLater(() -> {
                    setupRenderState();
                    mainWindow.zoomOutForBoundingBox(BoundingBox.of(polygons));
                    loadingDialog.hide();
                });
            });
            t.setUncaughtExceptionHandler((thread, e) -> {
                Platform.runLater(() -> loadingDialog.hide());
                logger.catching(e);
            });
            t.start();
        };
        refresher.run();
    }

    public void render(CityDoctorModel model) {
        errorUpdater = null;
        refresher = () -> {
            Platform.runLater(() -> {
                loadingDialog.show();
                clearGeometryTrees();
            });
            Thread t = new Thread(() -> {
                currentTriGeom = TriangulatedGeometry.of(model, lodFilters);
                errVisitor.setGeometry(currentTriGeom);
                Platform.runLater(() -> {
                    setupRenderState();
                    mainWindow.zoomOutForBoundingBox(BoundingBox.of(model));
                    loadingDialog.hide();
                });
            });
            t.setUncaughtExceptionHandler((thread, e) -> {
                Platform.runLater(() -> loadingDialog.hide());
                logger.catching(e);
            });
            t.start();
        };
        refresher.run();
    }

    public void renderVegetation(List<Vegetation> vegetation) {
        renderCityObjects(vegetation, Color.LIGHTGREEN);
    }

    public void renderTransportation(List<TransportationObject> transportation) {
        renderCityObjects(transportation, Color.YELLOW);
    }

    public void renderBridges(List<BridgeObject> bridges) {
        errorUpdater = null;
        refresher = () -> {
            Platform.runLater(() -> {
                loadingDialog.show();
                clearGeometryTrees();
            });
            Thread t = new Thread(() -> {
                Set<ConcretePolygon> polygons = new HashSet<>();
                for (BridgeObject b : bridges) {
                    collectBridgePolygons(polygons, b);
                    for (BridgeObject bp : b.getParts()) {
                        collectBridgePolygons(polygons, bp);
                    }
                }
                currentTriGeom = TriangulatedGeometry.of(polygons, Color.CORAL);
                errVisitor.setGeometry(currentTriGeom);
                Platform.runLater(() -> {
                    setupRenderState();
                    mainWindow.zoomOutForBoundingBox(BoundingBox.of(polygons));
                    loadingDialog.hide();
                });
            });
            t.setUncaughtExceptionHandler((thread, e) -> {
                Platform.runLater(() -> loadingDialog.hide());
                logger.catching(e);
            });
            t.start();
        };
        refresher.run();
    }

    public void renderWater(List<WaterObject> water) {
        renderCityObjects(water, Color.LIGHTSKYBLUE);
    }

    public void renderCityFurniture(List<CityFurniture> cityFurniture) {
        errorUpdater = null;
        refresher = () -> {
            Platform.runLater(() -> {
                loadingDialog.show();
                clearGeometryTrees();
            });
            Thread t = new Thread(() -> {
                Set<ConcretePolygon> polygons = new HashSet<>();
                for (CityFurniture cf : cityFurniture) {
                    collectCityFurniture(polygons, cf);
                }
                currentTriGeom = TriangulatedGeometry.of(polygons, Color.BLUEVIOLET);
                errVisitor.setGeometry(currentTriGeom);
                Platform.runLater(() -> {
                    setupRenderState();
                    mainWindow.zoomOutForBoundingBox(BoundingBox.of(polygons));
                    loadingDialog.hide();
                });
            });
            t.setUncaughtExceptionHandler((thread, e) -> {
                Platform.runLater(() -> loadingDialog.hide());
                logger.catching(e);
            });
            t.start();
        };
        refresher.run();

    }

    public void renderTerrain(List<CityObject> land) {
        renderLandObjects(land, Color.BROWN);
    }

    private void renderLandObjects(List<CityObject> cos, Color baseColor) {
        errorUpdater = null;
        refresher = () -> {
            Platform.runLater(() -> {
                loadingDialog.show();
                clearGeometryTrees();
            });
            Thread t = new Thread(() -> {
                Set<ConcretePolygon> polygons = new HashSet<>();
                for (CityObject co : cos) {
                    addPolygons(co, polygons);
                    if (co instanceof ReliefObject) {
                        ReliefObject relief = (ReliefObject) co;
                        for (TinObject tin : relief.getComponents()) {
                            addPolygons(tin, polygons);
                        }
                    }
                }
                currentTriGeom = TriangulatedGeometry.of(polygons, baseColor);
                errVisitor.setGeometry(currentTriGeom);
                Platform.runLater(() -> {
                    setupRenderState();
                    mainWindow.zoomOutForBoundingBox(BoundingBox.of(polygons));
                    loadingDialog.hide();
                });
            });
            t.setUncaughtExceptionHandler((thread, e) -> Platform.runLater(() -> loadingDialog.hide()));
            t.start();
        };
        refresher.run();
    }

    public void renderCityObjects(List<? extends CityObject> cos, Color baseColor) {
        errorUpdater = null;
        refresher = () -> {
            Platform.runLater(() -> {
                loadingDialog.show();
                clearGeometryTrees();
            });
            Thread t = new Thread(() -> {
                Set<ConcretePolygon> polygons = new HashSet<>();
                for (CityObject co : cos) {
                    addPolygons(co, polygons);
                }
                currentTriGeom = TriangulatedGeometry.of(polygons, baseColor);
                errVisitor.setGeometry(currentTriGeom);
                Platform.runLater(() -> {
                    setupRenderState();
                    mainWindow.zoomOutForBoundingBox(BoundingBox.of(polygons));
                    loadingDialog.hide();
                });
            });
            t.setUncaughtExceptionHandler((thread, e) -> Platform.runLater(() -> loadingDialog.hide()));
            t.start();
        };
        refresher.run();
    }

    public void renderTunnels(List<Tunnel> objects) {
        errorUpdater = null;
        refresher = () -> {
            Platform.runLater(() -> {
                loadingDialog.show();
                clearGeometryTrees();
            });
            Thread t = new Thread(() -> {
                Set<ConcretePolygon> polygons = new HashSet<>();
                for (Tunnel tunnel : objects) {
                    collectTunnelPolygons(polygons, tunnel);
                    for (TunnelPart tp : tunnel.getTunnelParts()) {
                        collectTunnelPolygons(polygons, tp);
                    }
                }
                currentTriGeom = TriangulatedGeometry.of(polygons);
                errVisitor.setGeometry(currentTriGeom);
                Platform.runLater(() -> {
                    setupRenderState();
                    mainWindow.zoomOutForBoundingBox(BoundingBox.of(polygons));
                    loadingDialog.hide();
                });
            });
            t.setUncaughtExceptionHandler((thread, e) -> {
                Platform.runLater(() -> loadingDialog.hide());
                logger.catching(e);
            });
            t.start();
        };
        refresher.run();
    }

    private void collectBuildingPolygons(Set<ConcretePolygon> polygons, AbstractBuilding ab) {
        addPolygons(ab, polygons);
        for (Installation bi : ab.getBuildingInstallations()) {
            addPolygons(bi, polygons);
            for (BoundarySurface bs : bi.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
                for (Opening o : bs.getOpenings()) {
                    addPolygons(o, polygons);
                }
            }
        }
        for (BuildingRoom br : ab.getBuildingRooms()) {
            addPolygons(br, polygons);
            for (BoundarySurface bs : br.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
            }
        }
        for (BoundarySurface bs : ab.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening o : bs.getOpenings()) {
                addPolygons(o, polygons);
            }
        }
    }

    private void collectBridgePolygons(Set<ConcretePolygon> polygons, BridgeObject br) {
        addPolygons(br, polygons);
        for (Installation bi : br.getBridgeInstallations()) {
            addPolygons(bi, polygons);
            for (BoundarySurface bs : bi.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
                for (Opening o : bs.getOpenings()) {
                    addPolygons(o, polygons);
                }
            }
        }
        for (BridgeRoom r : br.getBridgeRooms()) {
            addPolygons(r, polygons);
            for (BoundarySurface bs : r.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
            }
        }
        for (BridgeConstructiveElement bce : br.getConstructiveElements()) {
            addPolygons(bce, polygons);
            for (BoundarySurface bs : bce.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
            }
        }
        for (BoundarySurface bs : br.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening o : bs.getOpenings()) {
                addPolygons(o, polygons);
            }
        }

    }

    public void collectTunnelPolygons(Set<ConcretePolygon> polygons, AbstractTunnel at) {
        addPolygons(at, polygons);
        for (Installation ti : at.getTunnelInstallations()) {
            addPolygons(ti, polygons);
            for (BoundarySurface bs : ti.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
                for (Opening o : bs.getOpenings()) {
                    addPolygons(o, polygons);
                }
            }
        }
        for (TunnelHollow th : at.getTunnelHollows()) {
            addPolygons(th, polygons);
            for (BoundarySurface bs : th.getBoundarySurfaces()) {
                addPolygons(bs, polygons);
            }
        }
        for (BoundarySurface bs : at.getBoundarySurfaces()) {
            addPolygons(bs, polygons);
            for (Opening o : bs.getOpenings()) {
                addPolygons(o, polygons);
            }
        }
    }

    private void collectCityFurniture(Set<ConcretePolygon> polygons, CityFurniture cf) {
        addPolygons(cf, polygons);
    }

    private void addConcretePolygons(Set<ConcretePolygon> polygons, Geometry geom) {
        for (Polygon p : geom.getPolygons()) {
            if (p.getOriginal().getPartOfSurface() != null &&
                    !roofFilter.allowedToUse(p.getOriginal().getPartOfSurface(), p.getParent())) continue;
            if (p instanceof CompositePolygon) continue;
            polygons.add(p.getOriginal());
        }
    }

    private void setupRenderState() {
        currentTriGeom.setCullFace(currentCulling);
        currentTriGeom.setDrawMode(currentDrawMode);
        Platform.runLater(() -> {
            mainWindow.getMeshGroup().getChildren().clear();
            mainWindow.getMeshGroup().getChildren().addAll(currentTriGeom.getMeshes());
        });
        mainWindow.getGridButton().setDisable(false);
        mainWindow.getCullingButton().setDisable(false);
    }

    public void showWireFrame(boolean show) {
        if (currentTriGeom != null) {
            if (show) {
                currentDrawMode = DrawMode.LINE;
            } else {
                currentDrawMode = DrawMode.FILL;
            }
            currentTriGeom.setDrawMode(currentDrawMode);
        }
    }

    public void enableCulling(boolean enable) {
        if (currentTriGeom != null) {
            if (enable) {
                currentCulling = CullFace.BACK;
            } else {
                currentCulling = CullFace.NONE;
            }
            currentTriGeom.setCullFace(currentCulling);
        }
    }

    public void clearCurrentRender() {
        // don't render anything
        Platform.runLater(() -> {
            mainWindow.getMeshGroup().getChildren().clear();
            clearGeometryTrees();
        });
    }

    public void highlight(Polygon p) {
        highlightController.highlight(p, currentTriGeom);
    }

    public void highlight(LinearRing lr) {
        highlightController.highlight(lr, currentTriGeom);
    }

    public void highlight(Edge e) {
        highlightController.highlight(e, currentTriGeom);
    }

    public void highlight(Vertex v) {
        highlightController.highlight(v, currentTriGeom);
    }

    public void highlight(List<LinearRing> highlightedRings) {
        highlightController.highlight(highlightedRings, currentTriGeom);
    }

    public void highlightEdges(List<Edge> edges) {
        highlightController.highlightEdges(edges, currentTriGeom);
    }

    public void highlightPolygons(List<List<Polygon>> components) {
        highlightController.highlightPolygons(components, currentTriGeom);
    }

    public void addHighlight(Vertex vertex, Color c) {
        highlightController.addHighlight(vertex, currentTriGeom, c);
    }

    public void highlight(CheckError err) {
        err.accept(errVisitor);
    }

    public void refresh() {
        if (refresher != null) {
            refresher.run();
        }
    }

    public void updateErrors() {
        if (errorUpdater != null) {
            errorUpdater.run();
        }
    }

    public void clearHighlights() {
        highlightController.clearHighlights();
    }

    public void addHighlight(Polygon p) {
        highlightController.addHighlight(p, currentTriGeom);
    }

    public void addHighlight(Triangle3d t) {
        highlightController.highlight(t, currentTriGeom);
    }

    public void reset() {
        errorUpdater = null;
        refresher = null;
    }

}
